from pydantic import BaseModel
from typing import Union, List, Dict, Type, Optional
import logging

logger = logging.getLogger(__name__)


class TypeMapping:
    STRING_TO_TYPE: Dict[str, Type] = {
        "str": str,
        "int": int,
        "float": float,
        "bool": bool,
        "list": list
    }

    @classmethod
    def string_to_type(cls, type_string: str) -> Type:
        """Convert a string representation to its corresponding Python type."""
        return cls.STRING_TO_TYPE.get(type_string)


class Slot(BaseModel):
    name: str
    type: str
    value: Union[str, int, float, bool, List[str], None]
    enum: Optional[list[Union[str, int, float, bool, None]]]
    description: str
    prompt: str
    required: bool
    verified: bool

class SlotInput(BaseModel):
    name: str
    type: type
    value: Union[str, int, float, bool, List[str], None]
    enum: Optional[list[Union[str, int, float, bool, None]]]
    description: str

class SlotInputList(BaseModel):
    slot_input_list: list[SlotInput]

class SlotOutput(BaseModel):
    name: str
    value: Union[str, int, float, bool, List[str], None]

class SlotOutputGemini(BaseModel):
    name: str
    value: Union[str, int, float, bool, List[str]] = None

class SlotOutputList(BaseModel):
    slot_output_list: list[SlotOutput]

class SlotOutputListGemini(BaseModel):
    slot_output_list: list[SlotOutputGemini]

class Verification(BaseModel):
    thought: str
    verification_needed: bool


def format_slots(slots: list[dict]) -> list[Slot]:
    format_slots = []
    for slot in slots:
        format_slots.append(Slot(
            name=slot["name"],
            type=slot["type"],
            value=slot.get("value", None),
            enum=slot.get("enum", []),
            description=slot["description"],
            prompt=slot["prompt"],
            required=slot.get("required", False),
            verified=slot.get("verified", False)
        ))
    return format_slots


# format slots for slotfilling input
def format_slotfilling_input(slots: list[Slot]) -> SlotInputList:
    input_slots = []
    for slot in slots:
        input_slots.append(SlotInput(
            name=slot.name,
            type=TypeMapping.string_to_type(slot.type),
            value=slot.value,
            enum=slot.enum,
            description=slot.description
        ))
    return SlotInputList(slot_input_list=input_slots)


# format slots after slotfilling
def format_slotfilling_output(slots: list[Slot], response: SlotOutputList) -> list[Slot]:
    filled_slots = response.slot_output_list
    logger.info(f"filled_slots: {filled_slots}")
    for slot in slots:
        for filled_slot in filled_slots:
            if slot.name == filled_slot.name:
                slot.value = filled_slot.value
                break
    return slots



import logging

from langgraph.graph import StateGraph, START

from arklex.env.workers.worker import BaseWorker, register_worker
from arklex.utils.graph_state import MessageState, StatusEnum
from arklex.env.workers.utils.chat_client import ChatClient

from arklex.orchestrator.NLU.nlu import SlotFilling

logger = logging.getLogger(__name__)
# @register_worker
class HITLWorker(BaseWorker):  
    description = "This is a template for a HITL worker."
    mode = None
    params = None
    verifier = []
    
    slot_fill_api: SlotFilling = None
    
    # def __init__(self, server_ip: str, server_port: int, name: str):
    def __init__(self, **kwargs):
        super().__init__()
                
        self.action_graph = self._create_action_graph()

    def verify_literal(self, state: MessageState) -> tuple[bool, str]:
        """Override this method to allow verification on the message, either orchestrator's message or user's message
        Case: user's message
        Before the bot generate the response for the user's query, the framework decide whether it need to call human for the help because the user directly request so
        Case: orchestrator's message
        After the bot generate the response for the user's query, the framework decide whether it need to call human for the help because of the low confidence of the bot's response

        Args:
            message (str): _description_

        Returns:
            tuple[bool, str]: _description_
        """
        return True, ""
    
    def verify_slots(self, message) -> tuple[bool, str]:
        ''' Override this method to allow verification on the slots'''
        return True, ""
    
    def verify(self, state: MessageState) -> tuple[bool, str]:
        ''' Override this method to allow advanced verification on MessageState object'''
        need_hitl, message_literal = self.verify_literal(state)
        if need_hitl:
            return True, message_literal
        
        need_hitl, message_slot = self.verify_slots(state['slots'])
        if need_hitl:
            return True, message_slot
        
        return False, ""
    
    def init_slotfilling(self, slotsfillapi):
        self.slotfillapi = SlotFilling(slotsfillapi)
        
    def create_prompt(self):
        ''' Create a prompt for the HITL mc worker '''
        return self.params['intro'] + '\n' + '\n'.join(f"({key}) {item}" for key, item in self.params['choices'].items())
            
    def chat(self, state: MessageState) -> MessageState:
        ''' Connects to chat with the human in the loop '''
        client = ChatClient(self.server_ip, self.server_port, name=self.name, mode='c')
        return client.sync_main()
        
        # arklex pseudocode
        # chat_history = await server_chat() # BACKEND CHATS WITH USER HERE'''
        # state['messageFlow'] = to_message_flow(chat_history)
        # messageFlow['result'] = chat_history[-1]
        return state
    
    def multiple_choice(self, state: MessageState) -> MessageState:
        ''' Connects to give human in the loop multiple choice'''
        client = ChatClient(self.server_ip, self.server_port, name=self.name, mode='ro')
        return client.sync_main(message=self.create_prompt())
    
    def hitl(self, state: MessageState) -> str:
        ''' Human in the loop function '''
        result = None
        match self.mode:
            case "chat":
                chat_result = self.chat(state)
                state['user_message'].history += ('\n' + chat_result)
                state['user_message'].message = chat_result.split(f'{self.name}: ')[-1].split(':')[0]
                result = "Live Chat Completed"

            case "mc":
                attempts = self.params["max_retries"]
                
                for _ in range(attempts):
                    selection = self.multiple_choice(state)

                    result = self.params["choices"].get(selection)
                        
                    if result:
                        break
                else:
                    result = self.params['default']
            
            case _:
                return self.error(state)
                
        state['response'] = result
        return state
        
    def fallback(self, state: MessageState) -> MessageState:
        """The message of the fallback

        Args:
            state (MessageState): _description_

        Returns:
            : 
        """
        state['message_flow'] = "The user don't need human help"
        state['status'] = StatusEnum.COMPLETE.value
        return state
            
    def _create_action_graph(self):
        workflow = StateGraph(MessageState)
        # Add nodes for each worker
        workflow.add_node("hitl", self.hitl)
        # Add edges
        workflow.add_edge(START, "hitl")
        return workflow
        
    def execute(self, state: MessageState) -> MessageState:
        if not self.verify(state):
            return self.error(state)
        
        graph = self.action_graph.compile()
        result = graph.invoke(state)
        return result
    
@register_worker
class HITLWorkerTestChat(HITLWorker):
    """This worker is designed to start live chat locally
    Status: Not in use (as of 2025-02-20)

    Args:
        HITLWorker (_type_): _description_

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """

    description = "Chat with a real end user"
    mode = "chat"
    
    def __init__(self, **kwargs):
        super().__init__()
        self.server_ip = kwargs.get('server_ip')
        self.server_port = kwargs.get('server_port')
        self.name = kwargs.get('name')
        
        if not self.server_ip or not self.server_port:
            raise ValueError("Server IP and Port are required")
    
    def verify_literal(self, message: str) -> bool:
        return "chat" in message
    
@register_worker
class HITLWorkerTestMC(HITLWorker):
    """This worker is designed to start multiple choice human-in-the-loop worker locally
    Status: Not in use (as of 2025-02-20)

    Args:
        HITLWorker (_type_): _description_

    Returns:
        _type_: _description_
    """

    description = "Get confirmation from a real end user in purchasing"
    mode = "mc"
    params = {
        "intro": "Should the user continue with this purchase? (Y/N)",
        "max_retries": 5,
        'default': 'User is not allowed to continue with the purchase',
        "choices": {
            "Y": "User is allowed to continue with the purchase",
            "N": "User is not allowed to continue with the purchase"
        }
    }
    
    def __init__(self, **kwargs):
        super().__init__()
        self.server_ip = kwargs.get('server_ip')
        self.server_port = kwargs.get('server_port')
        self.name = kwargs.get('name')
    
    def verify_literal(self, message: str) -> bool:
        return "buy" in message
    
@register_worker
class HITLWorkerChatFlag(HITLWorker):
    """This worker is designed to start live chat with another built server. 
    So it will return the indicator of what type of human help needed.

    Args:
        HITLWorker (_type_): _description_

    Returns:
        MessageState: with hitl value in the MessageState[metadata]
    """
    description = "Human in the loop worker"
    mode = "chat"
    
    def verify_literal(self, state: MessageState) -> bool:
        """[TODO] Need to implement for orchestrator message as well
        (as of 2025-02-20)
        This method is to check the message from the user, since in the NLU, we already determine that the user wants to chat with the human in the loop.

        Args:
            message (str): _description_

        Returns:
            bool: _description_
        """
        message = "I'll connect you to a representative!"

        return True, message
    
    def execute(self, state: MessageState) -> MessageState:
        if not state['metadata'].get('hitl'):
            need_hitl, message = self.verify(state)
            if not need_hitl:
                return self.fallback(state)
            
            state["message_flow"] = message
            state['metadata']['hitl'] = 'live'
            state['status'] = StatusEnum.STAY.value
        
        else:
            state['message_flow'] = 'Live chat completed'
            state['metadata']['hitl'] = None
            state['status'] = StatusEnum.COMPLETE.value
        
        logger.info(state['message_flow'])
        return state
    
@register_worker
class HITLWorkerMCFlag(HITLWorker):
    """This worker is designed to start live chat with another built server. 
    So it will return the indicator of what type of human help needed.
    Status: Not in use (as of 2025-02-20)

    Args:
        HITLWorker (_type_): _description_

    Returns:
        MessageState: with hitl value in the MessageState[metadata]
    """

    description = "Get confirmation from a real end user in purchasing"
    mode = "mc"
    params = {
        "intro": "Should the user continue with this purchase? (Y/N)",
        "max_retries": 5,
        'default': 'User is not allowed to continue with the purchase',
        "choices": {
            "Y": "User is allowed to continue with the purchase",
            "N": "User is not allowed to continue with the purchase"
        }
    }
    
    def verify_literal(self, message: str) -> bool:
        return "buy" in message
    
    def execute(self, state: MessageState) -> MessageState:
        if not state['metadata'].get('hitl'):
            need_hitl, _ = self.verify(state)
            if not need_hitl:
                return self.fallback(state)
            
            state['response'] = '[[sending confirmation : this should not show up for user]]'
            state['metadata']['hitl'] = 'mc'
            state['metadata]']['attempts'] = self.params.get("max_retries", 3)
            state['status'] = StatusEnum.STAY.value
        
        else:
            result = self.params["choices"].get(state.user_message.message) # not actually user message but system confirmation
            
            if result:
                state['response'] = result
                state['metadata']['hitl'] = None
                state['status'] = StatusEnum.COMPLETE.value
                return state
            
            state['metadata]']['attempts'] -= 1
            if state['metadata]']['attempts'] <= 0:
                state['response'] = self.params['default']
                state['metadata']['hitl'] = None
                state['status'] = StatusEnum.INCOMPLETE.value
                return state
                    
            state['response'] = '[[sending confirmation : this should not show up for user]]'
            state['metadata']['hitl'] = 'mc'
            state['status'] = StatusEnum.STAY.value
            
        return state


from datetime import datetime, timezone

import hubspot
from hubspot.crm.objects.emails import PublicObjectSearchRequest, ApiException
from hubspot.crm.objects.communications.models import SimplePublicObjectInputForCreate
from hubspot.crm.associations.v4 import AssociationSpec

from arklex.env.tools.tools import register_tool, logger
from arklex.env.tools.hubspot.utils import authenticate_hubspot
from arklex.exceptions import ToolExecutionError

description = "Find the contacts record by email. If the record is found, the lastmodifieddate of the contact will be updated. If the correspodning record is not found, the function will return an error message."


slots = [
    {
        "name": "email",
        "type": "str",
        "description": "The email of the user, such as 'something@example.com'.",
        "prompt": "Thanks for your interest in our products! Could you please provide your email or phone number?",
        "required": True,
    },
    {
        "name": "chat",
        "type": "str",
        "description": "This occurs when user communicates with the chatbot",
        "prompt": "",
        "required": True,
    }
]
outputs = [
    {
        "name": "contact_information",
        "type": "str",
        "description": "The basic contact information for the existing customer (e.g. id, first_name, last_name, etc.)",
    }
]

USER_NOT_FOUND_PROMPT = "User not found (not an existing customer)"


@register_tool(description, slots, outputs)
def find_contact_by_email(email: str, chat: str, **kwargs) -> str:
    access_token = authenticate_hubspot(kwargs)

    api_client = hubspot.Client.create(access_token=access_token)
    public_object_search_request = PublicObjectSearchRequest(
        filter_groups=[
            {
                "filters": [
                    {
                        "propertyName": "email",
                        "operator": "EQ",
                        "value": email
                    }
                ]
            }
        ]
    )

    try:
        contact_search_response = api_client.crm.contacts.search_api.do_search(
            public_object_search_request=public_object_search_request)
        logger.info("Found contact by email: {}".format(email))
        contact_search_response = contact_search_response.to_dict()
        if contact_search_response['total'] == 1:
            contact_id = contact_search_response['results'][0]['id']
            communication_data = SimplePublicObjectInputForCreate(
                properties = {
                    "hs_communication_channel_type": "CUSTOM_CHANNEL_CONVERSATION",
                    "hs_communication_body": chat,
                    "hs_communication_logged_from": "CRM",
                    "hs_timestamp": datetime.now(timezone.utc).isoformat(),
                }
            )
            contact_info_properties = {
                'contact_id': contact_id,
                'contact_email': email,
                'contact_first_name': contact_search_response['results'][0]['properties'].get('firstname'),
                'contact_last_name': contact_search_response['results'][0]['properties'].get('lastname')
            }
            try:
                communication_creation_response = api_client.crm.objects.communications.basic_api.create(communication_data)
                communication_creation_response = communication_creation_response.to_dict()
                communication_id = communication_creation_response['id']
                association_spec = [
                    AssociationSpec(
                        association_category="HUBSPOT_DEFINED",
                        association_type_id=82
                    )
                ]
                try:
                    association_creation_response = api_client.crm.associations.v4.basic_api.create(
                        object_type="contact",
                        object_id=contact_id,
                        to_object_type="communication",
                        to_object_id=communication_id,
                        association_spec=association_spec
                    )
                except ApiException as e:
                    logger.info("Exception when calling AssociationV4: %s\n" % e)
            except ApiException as e:
                logger.info("Exception when calling basic_api: %s\n" % e)
            return str(contact_info_properties)
        else:
            raise ToolExecutionError(f"HubSpot find_contact_by_email failed: {e}", USER_NOT_FOUND_PROMPT)
    except ApiException as e:
        logger.info("Exception when calling search_api: %s\n" % e)
        raise ToolExecutionError(f"HubSpot find_contact_by_email failed: {e}", USER_NOT_FOUND_PROMPT)






from typing import Any, Dict
import json
import inspect
import shopify

from arklex.env.tools.tools import register_tool
from arklex.env.tools.shopify.utils import authorify_admin
from arklex.env.tools.shopify.utils_slots import ShopifyFindUserByEmailSlots, ShopifyOutputs
from arklex.exceptions import ToolExecutionError
from arklex.env.tools.shopify._exception_prompt import ExceptionPrompt

description = "Find user id by email. If the user is not found, the function will return an error message."
slots = ShopifyFindUserByEmailSlots.get_all_slots()
outputs = [
    ShopifyOutputs.USER_ID
]


@register_tool(description, slots, outputs)
def find_user_id_by_email(user_email: str, **kwargs) -> str:
    func_name = inspect.currentframe().f_code.co_name
    auth = authorify_admin(kwargs)
    if auth["error"]:
        return auth["error"]
    
    try:
        with shopify.Session.temp(**auth):
            response = shopify.GraphQL().execute(f"""
                {{
                    customers (first: 10, query: "email:{user_email}") {{
                        edges {{
                            node {{
                                id
                            }}
                        }}
                    }}
                }}
                """)
        nodes = json.loads(response)["data"]["customers"]["edges"]
        if len(nodes) == 1:
            user_id = nodes[0]["node"]["id"]
            return user_id
        else:
            raise ToolExecutionError(func_name, ExceptionPrompt.MULTIPLE_USERS_SAME_EMAIL_ERROR_PROMPT)
    except Exception as e:
        raise ToolExecutionError(func_name, ExceptionPrompt.USER_NOT_FOUND_ERROR_PROMPT)


import ast

import hubspot
from hubspot.crm.objects.emails import ApiException

from arklex.env.tools.tools import register_tool, logger
from arklex.env.tools.hubspot.utils import authenticate_hubspot
from arklex.exceptions import ToolExecutionError


description = "Find the owner id in the contact. If owner id is found, the next step is using the extracted owner id to find the information of the owner. "


slots = [
    {
        "name": "cus_cid",
        "type": "str",
        "description": "The id of the customer contact.",
        "prompt": "",
        "required": True,
    },
]

outputs = [
    {
        "name": "owner_id",
        "type": "str",
        "description": "The id of the owner of the contact.",
    }
]

OWNER_UNFOUND_PROMPT = 'Owner not found (not an existing customer)'


@register_tool(description, slots, outputs)
def find_owner_id_by_contact_id(cus_cid, **kwargs) -> str:
    access_token = authenticate_hubspot(kwargs)


    api_client = hubspot.Client.create(access_token=access_token)

    try:
        get_owner_id_response = api_client.api_request(
            {
                "path": "/crm/v3/objects/contacts/{}".format(cus_cid),
                "method": "GET",
                "headers": {
                    'Content-Type': 'application/json'
                },
                "qs": {
                    "properties": 'hubspot_owner_id'
                }
            }

        )
        get_owner_id_response = get_owner_id_response.json()

        owner_id = get_owner_id_response['properties']['hubspot_owner_id']

        return owner_id
    except ApiException as e:
        logger.info("Exception when extracting owner_id of one contact: %s\n" % e)
        raise ToolExecutionError(f"HubSpot find_owner_id_by_contact_id failed: {e}", OWNER_UNFOUND_PROMPT)



import json
import time
from typing import Any, Dict
import logging
import uuid
from typing import List, Dict, Any, Tuple
import copy
from arklex.env.env import Env
import janus
from dotenv import load_dotenv

from langchain_core.runnables import RunnableLambda

from arklex.orchestrator.task_graph import TaskGraph
from arklex.env.tools.utils import ToolGenerator
from arklex.types import EventType, StreamType
from arklex.utils.graph_state import (ConvoMessage, NodeInfo, OrchestratorMessage,
                                      MessageState, PathNode, StatusEnum,
                                      BotConfig, Slot, Timing, Metadata,
                                      Taskgraph, Memory, Params)
from arklex.utils.utils import format_chat_history


load_dotenv()
logger = logging.getLogger(__name__)

class AgentOrg:
    def __init__(self, config, env: Env, **kwargs):
        if isinstance(config, dict):
            self.product_kwargs = config
        else:
            self.product_kwargs = json.load(open(config))
        self.user_prefix = "user"
        self.worker_prefix = "assistant"
        self.environment_prefix = "tool"
        self.__eos_token = "\n"
        self.task_graph = TaskGraph("taskgraph", self.product_kwargs)
        self.env = env

    
    def init_params(self, inputs) -> Tuple[str, str, Params]:
        text = inputs["text"]
        chat_history = inputs["chat_history"]
        params = inputs["parameters"]

        timing = Timing(
            taskgraph=params.get("metadata", {}).get("timing", {}).get("taskgraph", None)
        )
        metadata = Metadata(
            chat_id=params.get("metadata", {}).get("chat_id", str(uuid.uuid4())),
            turn_id=params.get("metadata", {}).get("turn_id", 0),
            hitl=params.get("metadata", {}).get("hitl", None),
            timing=timing,
        )
        taskgraph = Taskgraph(
            dialog_states=params.get("taskgraph", {}).get("dialog_states", {}),
            path=params.get("taskgraph", {}).get("path", []),
            curr_node=params.get("taskgraph", {}).get("curr_node", None),
            curr_pred_intent=params.get("taskgraph", {}).get("curr_pred_intent", None),
            node_limit=params.get("taskgraph", {}).get("node_limit", {}),
            nlu_records=params.get("taskgraph", {}).get("nlu_records", []),
            node_status=params.get("taskgraph", {}).get("node_status", {}),
            available_global_intents=params.get("taskgraph", {}).get("available_global_intents", [])
        )
        memory = Memory(
            history=params.get("memory", {}).get("history", []),
            tool_response=params.get("memory", {}).get("tool_response", {}),
        )
        params = Params(
            metadata=metadata,
            taskgraph=taskgraph,
            memory=memory
        )

        chat_history_copy = copy.deepcopy(chat_history)
        chat_history_copy.append({"role": self.user_prefix, "content": text})
        chat_history_str = format_chat_history(chat_history_copy)
        if not params["taskgraph"]["dialog_states"]:
            params["taskgraph"]["dialog_states"] = {}
        params["metadata"]["turn_id"] += 1
        params["metadata"]["tool_response"] = {}
        if not params["memory"]["history"]:
            params["memory"]["history"] = copy.deepcopy(chat_history_copy)
        else:
            params["memory"]["history"].append(chat_history_copy[-2])
            params["memory"]["history"].append(chat_history_copy[-1])
        return text, chat_history_str, params

    def check_skip_node(self, node_info: NodeInfo, params: Params):
        if not node_info["can_skipped"]:
            return False
        cur_node_id = params["taskgraph"]["curr_node"]
        if cur_node_id in params["taskgraph"]["node_limit"]:
            if params["taskgraph"]["node_limit"][cur_node_id] <= 0:
                return True
        return False

    def post_process_node(self, node_info: NodeInfo, params: Params, update_info:dict={}):
        '''
        update_info is a dict of
            skipped = Optional[bool]
        '''
        curr_node = params["taskgraph"]["curr_node"]
        node = PathNode(
            node_id = curr_node,
            is_skipped = update_info.get("is_skipped", False),
            in_flow_stack=node_info.get("add_flow_stack", False),
            nested_graph_node_value = None,
            nested_graph_leaf_jump = None,
        )
        
        params["taskgraph"]['path'].append(node)

        if curr_node in params["taskgraph"]["node_limit"]:
            params["taskgraph"]["node_limit"][curr_node] -= 1
        return params
    
    def handl_direct_node(self, node_info: NodeInfo, params: Params):
        # Direct response
        node_attribute = node_info["attributes"]
        if node_attribute.get("value", "").strip() and node_attribute.get("direct"):
            return_response = {
                "answer": node_attribute["value"],
                "parameters": params
            }   
            if node_attribute.get("type", "") == "multiple-choice" and node_attribute.get("choice_list", []):
                return_response["choice_list"] = node_attribute["choice_list"]
            return True, return_response, params
        return False, None, params
    
    def perform_node(self, node_info: NodeInfo, params: Params, text: str, message_flow: str, chat_history_str: str, stream_type: StreamType, message_queue: janus.SyncQueue):
        # Tool/Worker
        user_message = ConvoMessage(history=chat_history_str, message=text)
        orchestrator_message = OrchestratorMessage(message=node_info["attributes"]["value"], attribute=node_info["attributes"])
        sys_instruct = "You are a " + self.product_kwargs["role"] + ". " + \
                            self.product_kwargs["user_objective"] + \
                            self.product_kwargs["builder_objective"] + \
                            self.product_kwargs["intro"] + \
                            self.product_kwargs.get("opt_instruct", "")
        bot_config = BotConfig(
            bot_id=self.product_kwargs.get("bot_id", "default"),
            version=self.product_kwargs.get("version", "default"),
            language=self.product_kwargs.get("language", "EN"),
            bot_type=self.product_kwargs.get("bot_type", "presalebot"),
        )
        message_state = MessageState(
            sys_instruct=sys_instruct, 
            bot_config=bot_config,
            user_message=user_message, 
            orchestrator_message=orchestrator_message, 
            trajectory=params["memory"]["history"], 
            message_flow=message_flow, 
            slots=params["taskgraph"]["dialog_states"],
            metadata=params["metadata"],
            is_stream=True if stream_type is not None else False,
            message_queue=message_queue
        )
        
        response_state, params = self.env.step(node_info["resource_id"], message_state, params)
        return response_state, params
    
    def get_response(self, inputs: dict, stream_type: StreamType = None, message_queue: janus.SyncQueue = None) -> Dict[str, Any]:
        text, chat_history_str, params = self.init_params(inputs)
        ##### TaskGraph Chain
        taskgraph_inputs = {
            "text": text,
            "chat_history_str": chat_history_str,
            "parameters": params,  ## TODO: different params for different components
            "allow_global_intent_switch": True, # ensure global intent changing happends only in the first get_node.
        }
        taskgraph_chain = RunnableLambda(self.task_graph.get_node) | RunnableLambda(self.task_graph.postprocess_node)

        
        counter_message_worker = 0
        counter_ragmsg_worker = 0
        counter_planner = 0 # TODO: when planner is re-implemented, execute/break the loop based on whether the planner should be used (bot config) .
        
        message_flow = "" # Store the message flow between different nodes

        n_node_performed = 0
        max_n_node_performed = 5
        while n_node_performed < max_n_node_performed:
            taskgraph_start_time = time.time()
            node_info, params = taskgraph_chain.invoke(taskgraph_inputs)
            taskgraph_inputs["allow_global_intent_switch"] = False
            params["metadata"]["timing"]["taskgraph"] = time.time() - taskgraph_start_time
            # Check if current node can be skipped
            can_skip = self.check_skip_node(node_info, params)
            if can_skip:
                params = self.post_process_node(node_info, params, {"is_skipped": True})
                continue
            logger.info(f"The current node info is : {node_info}")
            # Check current node attributes
            if node_info["resource_id"] == "planner":
                counter_planner += 1
            elif node_info["resource_id"] == self.env.name2id["MessageWorker"]:
                counter_message_worker += 1
            elif node_info["resource_id"] == self.env.name2id["RagMsgWorker"]:
                counter_ragmsg_worker += 1
            # handle direct node
            is_direct_node, direct_response, params = self.handl_direct_node(node_info, params)
            if is_direct_node:
                params = self.post_process_node(node_info, params)
                return direct_response
            # perform node
            response_state, params = self.perform_node(node_info,
                                                       params,
                                                       text,
                                                       message_flow,
                                                       chat_history_str,
                                                       stream_type,
                                                       message_queue)
            params = self.post_process_node(node_info, params)
            message_flow = response_state.get("message_flow", "")
            n_node_performed += 1

            # If the current node is not complete, then no need to continue to the next node
            node_status = params["taskgraph"]["node_status"]
            cur_node_id = params["taskgraph"]["curr_node"]
            status = node_status.get(cur_node_id, StatusEnum.COMPLETE.value)
            if status == StatusEnum.INCOMPLETE.value:
                break
            # If the counter of message worker or counter of default worker or counter of ragmsg worker == 1, break the loop
            if counter_message_worker == 1 or counter_planner == 1 or counter_ragmsg_worker == 1:
                break
            if node_info["is_leaf"] is True:
                break

        if not response_state.get("response", ""):
            logger.info("No response, do context generation")
            if stream_type is None:
                response_state = ToolGenerator.context_generate(response_state)
            else:
                response_state = ToolGenerator.stream_context_generate(response_state)
            params["memory"]["tool_response"] = {}
        
        response = response_state.get("response", "")
        return {
            "answer": response,
            "parameters": params,
            "human-in-the-loop": params['metadata'].get('hitl', None),
        }



import json
from typing import Any, Dict

import shopify

from arklex.env.tools.tools import register_tool
from arklex.env.tools.shopify.utils import authorify_admin
from arklex.env.tools.shopify.utils_slots import ShopifyGetOrderDetailsSlots, ShopifyOutputs
from arklex.exceptions import ToolExecutionError
from arklex.env.tools.shopify._exception_prompt import ExceptionPrompt
import inspect

description = "Get the status and details of an order."
slots = ShopifyGetOrderDetailsSlots.get_all_slots()
outputs = [
    ShopifyOutputs.ORDERS_DETAILS
]


@register_tool(description, slots, outputs)
def get_order_details(order_ids: list, order_names: list, user_id: str, limit=10, **kwargs) -> str:
    func_name = inspect.currentframe().f_code.co_name
    limit = int(limit) if limit else 10
    auth = authorify_admin(kwargs)
    
    try:
        query = f"customer_id:{user_id.split('/')[-1]}"
        if order_ids:
            order_ids = ' OR '.join(f'id:{oid.split("/")[-1]}' for oid in order_ids)
            query += f" AND ({order_ids})"
        if order_names:
            order_names = ' OR '.join(f'name:{name}' for name in order_names)
            query += f" AND ({order_names})"
        with shopify.Session.temp(**auth):
            response = shopify.GraphQL().execute(f"""
            {{
                orders (first: {limit}, query:"{query}") {{
                    nodes {{
                        id
                        name
                        createdAt
                        cancelledAt
                        returnStatus
                        statusPageUrl
                        totalPriceSet {{
                            presentmentMoney {{
                                amount
                            }}
                        }}
                        fulfillments {{
                            displayStatus
                            trackingInfo {{
                                number
                                url
                            }}
                        }}
                        lineItems(first: 10) {{
                            edges {{
                                node {{
                                    id
                                    title
                                    quantity
                                    variant {{
                                        id
                                        product {{
                                            id
                                        }}
                                    }}
                                }}
                            }}
                        }}
                    }}
                }}
            }}
            """)
            result = json.loads(response)["data"]["orders"]["nodes"]
            if len(result) == 0:
                return "You have no orders placed."
            response_text = ""
            for order in result:
                response_text += f"Order ID: {order.get('id', 'None')}\n"
                response_text += f"Order Name: {order.get('name', 'None')}\n"
                response_text += f"Created At: {order.get('createdAt', 'None')}\n"
                response_text += f"Cancelled At: {order.get('cancelledAt', 'None')}\n"
                response_text += f"Return Status: {order.get('returnStatus', 'None')}\n"
                response_text += f"Status Page URL: {order.get('statusPageUrl', 'None')}\n"
                response_text += f"Total Price: {order.get('totalPriceSet', {}).get('presentmentMoney', {}).get('amount', 'None')}\n"
                response_text += f"Fulfillment Status: {order.get('fulfillments', 'None')}\n"
                response_text += "Line Items:\n"
                for item in order.get('lineItems', {}).get('edges', []):
                    response_text += f"    Title: {item.get('node', {}).get('title', 'None')}\n"
                    response_text += f"    Quantity: {item.get('node', {}).get('quantity', 'None')}\n"
                    response_text += f"    Variant: {item.get('node', {}).get('variant', {})}\n"
                response_text += "\n"
        return response_text
    except Exception as e:
        raise ToolExecutionError(func_name, ExceptionPrompt.ORDERS_NOT_FOUND_PROMPT)



from langchain_openai import ChatOpenAI
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_anthropic import ChatAnthropic
from langchain_openai import OpenAIEmbeddings
from langchain_google_genai import GoogleGenerativeAIEmbeddings
from langchain_huggingface.embeddings import HuggingFaceEmbeddings
from langchain_huggingface import HuggingFaceEndpoint,ChatHuggingFace


def get_huggingface_llm(model, **kwargs):
    llm = HuggingFaceEndpoint(
        repo_id=model,
        task="text-generation",
        **kwargs
    )
    return ChatHuggingFace(llm=llm)

LLM_PROVIDERS = ["openai", "gemini", "anthropic", "huggingface"]

PROVIDER_MAP = {
    "anthropic": ChatAnthropic,
    "gemini": ChatGoogleGenerativeAI,
    "openai": ChatOpenAI,
    "huggingface": get_huggingface_llm
}

PROVIDER_EMBEDDINGS = {
    "anthropic": HuggingFaceEmbeddings,
    "gemini": GoogleGenerativeAIEmbeddings,
    "openai": OpenAIEmbeddings ,
    "huggingface": HuggingFaceEmbeddings
}
PROVIDER_EMBEDDING_MODELS = {
    "anthropic": "sentence-transformers/all-mpnet-base-v2",
    "gemini": "models/embedding-001",
    "openai": "text-embedding-ada-002",
    "huggingface": "sentence-transformers/all-mpnet-base-v2",
}


from typing import TypedDict, Any, Annotated, Optional
from arklex.utils.slot import Slot
import janus
from pydantic import BaseModel
from enum import Enum


### Bot-related classes
class BotConfig(BaseModel):
    bot_id: str
    version: str
    language: str
    bot_type: str

### Message-related classes

class ConvoMessage(BaseModel):
    history: str # it could be the whole original message or the summarization of the previous conversation from memory module
    message: str


class OrchestratorMessage(BaseModel):
    message: str
    attribute: dict


### Task status-related classes

class StatusEnum(Enum):
    COMPLETE = "complete"
    INCOMPLETE = "incomplete"
    STAY = "stay"


class MessageState(TypedDict):
    # system configuration
    sys_instruct: str
    # bot configuration
    bot_config: BotConfig
    # input message
    user_message: ConvoMessage
    orchestrator_message: OrchestratorMessage
    # action trajectory
    trajectory: list[dict[str, any]]
    # message flow between different nodes
    message_flow: Annotated[str, "message flow between different nodes"]
    # final response
    response: str
    # task-related params
    status: StatusEnum
    slots: dict[str, list[Slot]]  # record the dialogue states of each action
    metadata: dict[str, any]
    # stream
    is_stream: bool
    message_queue: janus.SyncQueue


class Timing(TypedDict):
    taskgraph: Optional[float]

class PathNode(TypedDict):
    node_id: str
    is_skipped: bool
    in_flow_stack: bool
    nested_graph_node_value: Optional[str]
    nested_graph_leaf_jump: Optional[str]


class Metadata(TypedDict):
    chat_id: str
    turn_id: int
    hitl: Optional[str]
    timing: Timing

class Taskgraph(TypedDict):
    dialog_states: dict[str, Any]
    path: list[PathNode]
    curr_node: Optional[str]
    intent: Optional[str]
    node_limit: dict[str, int]
    nlu_records: list
    node_status: dict[str, StatusEnum]
    available_global_intents: list

class Memory(TypedDict):
    history: list[dict[str, Any]]
    tool_response: dict

class Params(TypedDict):
    metadata: Metadata
    taskgraph: Taskgraph
    memory: Memory

class NodeInfo(TypedDict):
    resource_id: str
    resource_name: str
    can_skipped: bool
    is_leaf: bool
    attributes: dict[str, Any]
    add_flow_stack: Optional[bool]


from datetime import datetime

import hubspot
import parsedatetime
from hubspot.crm.objects.meetings import ApiException

from arklex.env.tools.tools import register_tool, logger
from arklex.env.tools.hubspot.utils import authenticate_hubspot
from arklex.exceptions import ToolExecutionError

description = "Give the customer that the unavailable time of the specific representative and the representative's related meeting link information."

slots = [
    {
        "name": "owner_id",
        "type": "str",
        "description": "The owner id of the owner.'",
        "prompt": "",
        "required": True,
    },
    {
        "name": "time_zone",
        "type": "str",
        "enum": ["America/New_York", "America/Los_Angeles", "Asia/Tokyo", "Europe/London"],
        "description": "The timezone of the user. For example, 'America/New_York'. If you are not sure, just ask the user to confirm.",
        "prompt": "Could you please provide your timezone or where are you now?",
        "required": True
    },
    {
        "name": "meeting_date",
        "type": "str",
        "description": "The exact date the customer want to take meeting with the representative. e.g. today, Next Monday, May 1st.",
        "prompt": "Could you please give me the date of the meeting?",
        "required": True,
    }
]
outputs = [
    {
        "name": "meeting_info",
        "type": "string",
        "description": "The unavailable time slots of the representative and the corresponding slug.",
    }
]

MEETING_LINK_UNFOUND_PROMPT = "The representative does not have a meeting link."


@register_tool(description, slots, outputs)
def check_available(owner_id: str, time_zone: str, meeting_date: str, **kwargs) -> str:
    access_token = authenticate_hubspot(kwargs)
    api_client = hubspot.Client.create(access_token=access_token)
    meeting_info = {
        'busy_time_slots': [],
        'busy_time_slots_unix': []
    }
    try:
        meeting_link_response = api_client.api_request(
            {
                "path": "/scheduler/v3/meetings/meeting-links",
                "method": "GET",
                "headers": {
                    'Content-Type': 'application/json'
                },
                "qs": {
                    'organizerUserId': owner_id
                }
            }
        )
        meeting_link_response = meeting_link_response.json()

        if meeting_link_response.get('total') == 0:
            raise ToolExecutionError("HubSpot check_available failed", MEETING_LINK_UNFOUND_PROMPT)
        else:
            meeting_links = meeting_link_response['results'][0]
        meeting_slug = meeting_links['slug']
        meeting_info['slug'] = meeting_slug
        try:
            availability_response = api_client.api_request(
                {
                    "path": "/scheduler/v3/meetings/meeting-links/book/{}".format(meeting_slug),
                    "method": "GET",
                    "headers": {
                        'Content-Type': 'application/json'
                    },
                    "qs": {
                        'timezone': time_zone
                    }
                }
            )
            cal = parsedatetime.Calendar()
            time_struct, _ = cal.parse(meeting_date)
            meeting_date = datetime(*time_struct[:3])
            availability_response = availability_response.json()
            busy_times = availability_response['allUsersBusyTimes'][0]['busyTimes']
            for busy_time in busy_times:
                start_time = datetime.fromtimestamp(busy_time["start"] / 1000)
                end_time = datetime.fromtimestamp(busy_time["end"] / 1000)

                if start_time.date() == meeting_date.date():
                    meeting_info['busy_time_slots'].append({
                        "start": start_time.isoformat(),
                        "end": end_time.isoformat()
                    })
                    meeting_info['busy_time_slots_unix'].append({
                        "start": busy_time["start"],
                        "end": busy_time["end"]
                    })
            return str(meeting_info)
        except ApiException as e:
            logger.info("Exception when extracting booking information of someone: %s\n" % e)
            raise ToolExecutionError(f"HubSpot check_available failed: {e}", MEETING_LINK_UNFOUND_PROMPT)
    except ApiException as e:
        logger.info("Exception when extracting meeting scheduler links: %s\n" % e)
        raise ToolExecutionError(f"HubSpot check_available failed: {e}", MEETING_LINK_UNFOUND_PROMPT)







__all__ = [
    "AuthenticationError",
    "ToolExecutionError",
]


class AuthenticationError(Exception):
    """
    Exception raised when authentication fails.
    """
    def __init__(self, message: str):
        self.message = f"Authentication failed: {message}"
        super().__init__(self.message)


class UserFacingError(Exception):
    """
    Exception raised to guide the user to update their query.
    """
    def __init__(self, message: str, extra_message: str):
        super().__init__(message)
        # Store the additional message in a custom attribute, which will be used to guide the user to update their query.
        self.extra_message = extra_message


class ToolExecutionError(UserFacingError):
    """
    Exception raised when a tool execution fails.
    """
    def __init__(self, message: str, extra_message: str):
        self.message = f"Tool {message} execution failed"
        super().__init__(self.message, extra_message)



"""
This module is currently inactive.

It is reserved for future use and may contain experimental or planned features.

Status:
    - Not in use (as of 2025-02-20)
    - Intended for future feature expansion

Module Name: chat_client

This file contains the code for setting up a chat client that can connect to a chat server and send and receive messages.
"""

import sys
import asyncio
import json
import argparse

import threading
from concurrent.futures import Future

# Shared future for async result
async_result = Future()

class ChatClient:
    ''' Chat client class '''
    reader, writer = None, None
    
    def __init__(self, server_address='127.0.0.1', server_port='8888', *, name=None, mode='c', debug=False):
        self.server_address = server_address
        self.server_port = server_port
        self.name = name
        self.mode = mode
        
        self.timeout = 100000
        self.debug = debug
        
        self.logs = []
        
        self.async_thread = None
        self.async_result = Future()
    
    # Thread for async event loop
    def async_thread_worker(self, message: str):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        
        result = loop.run_until_complete(self.main(message))
        self.async_result.set_result(result)
        
    def sync_main(self, message: str=None) -> str:
        self.async_thread = threading.Thread(target=self.async_thread_worker, args=(message,))
        self.async_thread.start()
        return self.async_result.result()
    
    def format_logs(self, logs: list[dict]) -> str:
        return '\n'.join([f"{log['name']}: {log['message']}" for log in logs])
        
    async def send_message(self, message):
        msg_bytes = json.dumps({'name': self.name, 'message': message}).encode()
        # transmit the message to the server
        self.writer.write(msg_bytes)
        # wait for the buffer to be empty
        await self.writer.drain()
            
        
    # send message to server
    async def write_messages(self):
        # read messages from the user and transmit to server
        while True:
            # read from stdin
            message = await asyncio.to_thread(sys.stdin.readline)
            
            # clear the input line
            CURSOR_UP = '\033[F'
            ERASE_LINE = '\033[K'
            print(CURSOR_UP + ERASE_LINE + CURSOR_UP)
            sys.stdout.flush()
            
            await self.send_message(message)
                
            # check if the user wants to quit the program
            if message.strip() == 'QUIT':
                # exit the loop
                break
            
        # report that the program is terminating
        print('Quitting...')
        
    async def receive_message(self):
        result_bytes = await self.reader.read(1024)  # Read raw bytes
        decoded_data = result_bytes.decode()
        messages = []
        # print(decoded_data)
        for data in decoded_data.split('{'):
            if not data:
                continue
            
            # convert to string
            # data_json = json.loads('{' + data)
            # name, line = data_json['name'], data_json['message'].strip()
            result_json = json.loads('{' + data)
            if self.debug or (self.mode == 'c' and result_json['name'] != 'Server'):
                print(f"{result_json['name']}: {result_json['message'].strip()}")
            messages.append(result_json)
            
        return messages
        
    # read messages from the server
    async def read_messages(self):
        # read messages from the server and print to user
        while True:
            messages = await self.receive_message()
            self.logs.extend(messages)
    
    # echo client
    async def main(self, message=None):
        # report progress to the user
        if self.debug:
            print(f'Connecting to {self.server_address}:{self.server_port}...')
        # open a connection to the server
        self.reader, self.writer = await asyncio.open_connection(self.server_address, self.server_port)
        # report progress to the user
        if self.debug:
            print('Connected.')
        
        if not self.name:
            self.name = input("Enter name: ")
        
        connect = {"sys": "connect", "name": self.name}
        connect_json = json.dumps(connect)
        self.writer.write(connect_json.encode())
        await self.writer.drain()
        
        await self.receive_message()
        await self.receive_message()
        
        read_task, write_task = None, None
        
        match self.mode:
            case 'c':
                # read and report messages from the server
                read_task = asyncio.create_task(self.read_messages())
                
                # write messages to the server
                write_task = asyncio.create_task(self.write_messages())
                    
                # wait for either task to complete
                done, pending = await asyncio.wait(
                    [read_task, write_task],
                    return_when=asyncio.FIRST_COMPLETED,
                )
                
                # cancel the remaining tasks
                for task in pending:
                    task.cancel()
                    
                await self.writer.drain()
                # report progress to the user
                print('Disconnecting from server...')
                # close the stream writer
                self.writer.close()
                # wait for the tcp connection to close
                await self.writer.wait_closed()
                # report progress to the user
                print('Done.')
                return self.format_logs(self.logs)

            case 'ro':
                if message: 
                    await self.send_message(message)
                read_msg = None
                while not read_msg:
                    for m in await self.receive_message():
                        if m['name'] != 'Server' and m['name'] != self.name:
                            read_msg = m['message']
                await self.send_message("QUIT")
                return read_msg
            
            case 'wo':
                await self.send_message(message)
                await self.send_message("QUIT")
                return None
                
            case _:
                raise ValueError(f"Invalid mode: {self.mode}")
            
        
        # report progress to the user
        print('Disconnecting from server...')
        # close the stream writer
        self.writer.close()
        # wait for the tcp connection to close
        await self.writer.wait_closed()
        # report progress to the user
        print('Done.')
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--server-address', type=str, default='127.0.0.1')
    parser.add_argument('--server-port', type=int, default=8888)
    parser.add_argument('--name', type=str, default='User')
    parser.add_argument('--timeout', type=int, default=100000)
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--mode', default='c', choices=['c', 'ro', 'wo'])
    parser.add_argument('--message', type=str, default=None)
    args = parser.parse_args()
    
    # create the chat client
    client = ChatClient(
                server_address=args.server_address, 
                server_port=args.server_port, 
                name=args.name, 
                mode=args.mode, 
                debug=args.debug
            )
    # run the event loop
    asyncio.run(client.main(args.message))
    


import copy
import logging
import collections
from typing import Tuple

import networkx as nx
import numpy as np
from langchain_openai import ChatOpenAI


from arklex.utils.model_provider_config import PROVIDER_MAP
from arklex.utils.utils import normalize, str_similarity, format_chat_history
from arklex.utils.graph_state import NodeInfo, Params, StatusEnum
from arklex.orchestrator.NLU.nlu import NLU, SlotFilling
from arklex.utils.model_config import MODEL

logger = logging.getLogger(__name__)

class TaskGraphBase:
    def __init__(self, name, product_kwargs):
        self.graph = nx.DiGraph(name=name)
        self.product_kwargs = product_kwargs
        self.create_graph()
        self.intents = self.get_pred_intents() # global intents
        self.start_node = self.get_start_node()

    def create_graph(self):
        raise NotImplementedError

    def get_pred_intents(self):
        intents = collections.defaultdict(list)
        for edge in self.graph.edges.data():
            if edge[2].get("attribute", {}).get("pred", False):
                edge_info = copy.deepcopy(edge[2])
                edge_info["source_node"] = edge[0]
                edge_info["target_node"] = edge[1]
                intents[edge[2].get("intent")].append(edge_info)
        return intents
    
    def get_start_node(self):
        for node in self.graph.nodes.data():
            if node[1].get("type", "") == "start":
                return node[0]
        return None


class TaskGraph(TaskGraphBase):
    def __init__(self, name: str, product_kwargs: dict):
        super().__init__(name, product_kwargs)
        self.unsure_intent = {
                "intent": "others",
                "source_node": None,
                "target_node": None,
                "attribute": {
                    "weight": 1,
                    "pred": False,
                    "definition": "",
                    "sample_utterances": []
                }
            }
        self.initial_node = self.get_initial_flow()
        self.model = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
            model=MODEL["model_type_or_path"], timeout=30000
        )
        self.nluapi = NLU(self.product_kwargs.get("nluapi"))
        self.slotfillapi = SlotFilling(self.product_kwargs.get("slotfillapi"))

    def create_graph(self):
        nodes = self.product_kwargs["nodes"]
        edges = self.product_kwargs["edges"]
        # convert the intent into lowercase
        for edge in edges:
            edge[2]['intent'] = edge[2]['intent'].lower()
        self.graph.add_nodes_from(nodes)
        self.graph.add_edges_from(edges)

    def get_initial_flow(self):
        services_nodes = self.product_kwargs.get("services_nodes", None)
        node = None
        if services_nodes:
            candidates_nodes = [v for k, v in services_nodes.items()]
            candidates_nodes_weights = [list(self.graph.in_edges(n, data="attribute"))[0][2]["weight"] for n in candidates_nodes]
            node = np.random.choice(candidates_nodes, p=normalize(candidates_nodes_weights))
        return node

    def jump_to_node(self, pred_intent, intent_idx, curr_node):
        """
        Jump to a node based on the intent
        """
        logger.info(f"pred_intent in jump_to_node is {pred_intent}")
        try:
            candidates_nodes = [self.intents[pred_intent][intent_idx]]
            candidates_nodes_weights = [node["attribute"]["weight"] for node in candidates_nodes]
            if candidates_nodes:
                next_node = np.random.choice([node["target_node"] for node in candidates_nodes], p=normalize(candidates_nodes_weights))
                next_intent = pred_intent
            else:  # This is for protection, logically shouldn't enter this branch
                next_node = curr_node
                next_intent = list(self.graph.in_edges(curr_node, data="intent"))[0][2]
        except Exception as e:
            logger.error(f"Error in jump_to_node: {e}")
            next_node = curr_node
            next_intent = list(self.graph.in_edges(curr_node, data="intent"))[0][2]
        return next_node, next_intent
    
    def _get_node(self, sample_node, params: Params, intent=None) -> Tuple[NodeInfo, Params]:
        """
        Get the output format (NodeInfo, Params) that get_node should return
        """
        logger.info(f"available_intents in _get_node: {params['taskgraph']['available_global_intents']}")
        logger.info(f"intent in _get_node: {intent}")
        node_info = self.graph.nodes[sample_node]
        resource_name = node_info["resource"]["name"]
        resource_id = node_info["resource"]["id"]
        if intent and intent in params["taskgraph"]["available_global_intents"]:
            # delete the corresponding node item from the intent list
            for item in params["taskgraph"]["available_global_intents"].get(intent, []):
                if item["target_node"] == sample_node:
                    params["taskgraph"]["available_global_intents"][intent].remove(item)
            if not params["taskgraph"]["available_global_intents"][intent]:
                params["taskgraph"]["available_global_intents"].pop(intent)
        
        params["taskgraph"]["curr_node"] = sample_node
        
        node_info = NodeInfo(
            resource_id=resource_id,
            resource_name=resource_name,
            can_skipped=True,
            is_leaf=len(list(self.graph.successors(sample_node))) == 0,
            attributes=node_info["attribute"],
            add_flow_stack=False
        )
        
        return node_info, params

    def _postprocess_intent(self, pred_intent, available_global_intents):
        found_pred_in_avil = False
        real_intent = pred_intent
        idx = 0
        # check whether there are __<{idx}> in the pred_intent
        if "__<" in pred_intent:
            real_intent = pred_intent.split("__<")[0]
        # get the idx
            idx = int(pred_intent.split("__<")[1].split(">")[0])
        for item in available_global_intents:
            if str_similarity(real_intent, item) > 0.9:
                found_pred_in_avil = True
                real_intent = item
                break
        return found_pred_in_avil, real_intent, idx

    def get_current_node(self, params: Params):
        """
        Get current node
        If current node is unknown, use start node
        """
        curr_node = params["taskgraph"].get("curr_node", None)
        if not curr_node or curr_node not in self.graph.nodes:
            curr_node = self.start_node
            params["taskgraph"]["intent"] = None
        else:
            curr_node = str(curr_node)
        params["taskgraph"]["curr_node"] = curr_node
        return curr_node, params
    
    def get_available_global_intents(self, params: Params):
        """
        Get available global intents
        """
        available_global_intents = params["taskgraph"].get("available_global_intents", [])
        if not available_global_intents:
            available_global_intents = copy.deepcopy(self.intents)
            if self.unsure_intent.get("intent") not in available_global_intents.keys():
                available_global_intents[self.unsure_intent.get("intent")].append(self.unsure_intent)
        logger.info(f"Available global intents: {available_global_intents}")
        return available_global_intents
    
    def update_node_limit(self, params: Params):
        """
        Update the node_limit in params which will be used to check if we can skip the node or not
        """
        old_node_limit = params["taskgraph"].get("node_limit", {})
        node_limit = {}
        for node in self.graph.nodes.data():
            limit = old_node_limit.get(node[0], node[1].get("limit")) 
            if limit is not None:
                node_limit[node[0]] = limit
        params["taskgraph"]["node_limit"] = node_limit
        return params

    def get_local_intent(self, curr_node, params: Params):
        """
        Get the local intent of a current node
        """
        candidates_intents = collections.defaultdict(list)
        for u, v, data in self.graph.out_edges(curr_node, data=True):
            intent = data.get("intent")
            if intent != "none" and data.get("intent"):
                edge_info = copy.deepcopy(data)
                edge_info["source_node"] = u
                edge_info["target_node"] = v
                candidates_intents[intent].append(edge_info)
        logger.info(f"Current local intent: {candidates_intents}")
        return dict(candidates_intents)

    
    def get_last_flow_stack_node(self, params: Params):
        """
        Get the last flow stack node from path
        """
        path = params["taskgraph"]["path"]
        for i in range(len(path) - 1, -1, -1):
            if path[i]["in_flow_stack"]:
                params["taskgraph"]["path"][i]["in_flow_stack"] = False
                return params["taskgraph"]["path"][i]["node_id"]
        return None
    
    def handle_multi_step_node(self, curr_node, params: Params) -> Tuple[bool, dict, Params]:
        """
        In case of a node having status == STAY, returned directly the same node
        """
        node_status = params["taskgraph"].get("node_status", {})
        logger.info(f"node_status: {node_status}")
        status = node_status.get(curr_node, StatusEnum.COMPLETE.value)
        if status == StatusEnum.STAY.value:
            node_info = self.graph.nodes[curr_node]
            node_name = node_info["resource"]["name"]
            id = node_info["resource"]["id"]
            node_output = {"id": id, "name": node_name, "attribute": node_info["attribute"]}
            return True, node_output, params
        return False, {}, params
    
    def handle_incomplete_node(self, curr_node, params: Params) -> Tuple[bool, dict, Params]:
        """
        If node is incomplete, return directly the node
        """
        node_status = params["taskgraph"].get("node_status", {})
        status = node_status.get(curr_node, StatusEnum.COMPLETE.value)
        if status == StatusEnum.INCOMPLETE.value:
            logger.info(f"no local or global intent found, the current node is not complete")
            node_info, params = self._get_node(curr_node, params)
            return True, node_info, params
        return False, {}, params
    
    def global_intent_prediction(self, curr_node, params: Params, available_global_intents, excluded_intents) -> Tuple[bool, str, dict, Params]:
        """
        Do global intent prediction
        """
        candidate_intents = copy.deepcopy(available_global_intents)
        candidate_intents = {k: v for k, v in candidate_intents.items() if k not in excluded_intents}
        pred_intent = None
        # if only unsure_intent is available -> move directly to this intent
        if len(candidate_intents) == 1 and self.unsure_intent.get("intent") in candidate_intents.keys():
            pred_intent = self.unsure_intent.get("intent")
        else: # global intent prediction
            # if match other intent, add flow, jump over
            candidate_intents[self.unsure_intent.get("intent")] = \
                candidate_intents.get(self.unsure_intent.get("intent"), [self.unsure_intent])
            logger.info(f"Available global intents with unsure intent: {candidate_intents}")
            
            pred_intent = self.nluapi.execute(self.text, candidate_intents, self.chat_history_str)
            params["taskgraph"]["nlu_records"].append({"candidate_intents": candidate_intents, 
                                "pred_intent": pred_intent, "no_intent": False, "global_intent": True})
            found_pred_in_avil, pred_intent, intent_idx = self._postprocess_intent(pred_intent, available_global_intents)
            # if found prediction and prediction is not unsure intent and current intent
            # TODO: how to know if user want to proceed or going back to the initial node of the same global intent
            # TODO: use pred_intent != params["taskgraph"]["global_intent"], but global_intent is not stored now.
            if found_pred_in_avil and pred_intent != self.unsure_intent.get("intent"):
                params["taskgraph"]["intent"] = pred_intent
                next_node, next_intent = self.jump_to_node(pred_intent, intent_idx, curr_node)
                logger.info(f"curr_node: {next_node}")
                node_info, params = self._get_node(next_node, params, intent=next_intent)
                # if current node is not a leaf node and jump to another node, then add it onto stack
                if next_node != curr_node and list(self.graph.successors(curr_node)):
                    node_info["add_flow_stack"] = True
                return True, pred_intent, node_info, params
        return False, pred_intent, {}, params

    
    def handle_random_next_node(self, curr_node, params: Params) -> Tuple[bool, dict, Params]:
        candidate_samples = []
        candidates_nodes_weights = []
        for out_edge in self.graph.out_edges(curr_node, data=True):
            if out_edge[2]["intent"] == "none":
                candidate_samples.append(out_edge[1])
                candidates_nodes_weights.append(out_edge[2]["attribute"]["weight"])
        if candidate_samples:
            # randomly choose one sample from candidate samples
            next_node = np.random.choice(candidate_samples, p=normalize(candidates_nodes_weights))
        else:  # leaf node + the node without None intents
            next_node = curr_node

        if next_node != curr_node:  # continue if curr_node is not leaf node, i.e. there is a actual next_node
            logger.info(f"curr_node: {next_node}")
            node_info, params = self._get_node(next_node, params)
            if params.get("nlu_records", None):
                params["nlu_records"][-1]["no_intent"] = True  # move on to the next node
            else: # only others available
                params["nlu_records"] = [{"candidate_intents": [], "pred_intent": "", "no_intent": True, "global_intent": False}]
            return True, node_info, params
        return False, {}, params
    
    def local_intent_prediction(self, curr_node, params: Params, curr_local_intents) -> Tuple[bool, dict, Params]:
        """
        Do local intent prediction
        """
        curr_local_intents_w_unsure = copy.deepcopy(curr_local_intents)
        curr_local_intents_w_unsure[self.unsure_intent.get("intent")] = \
            curr_local_intents_w_unsure.get(self.unsure_intent.get("intent"), [self.unsure_intent])
        logger.info(f"Check intent under current node: {curr_local_intents_w_unsure}")
        pred_intent = self.nluapi.execute(self.text, curr_local_intents_w_unsure, self.chat_history_str)
        params["taskgraph"]["nlu_records"].append({"candidate_intents": curr_local_intents_w_unsure, 
                                "pred_intent": pred_intent, "no_intent": False, "global_intent": False})
        found_pred_in_avil, pred_intent, intent_idx = self._postprocess_intent(pred_intent, curr_local_intents)
        logger.info(f"Local intent predition -> found_pred_in_avil: {found_pred_in_avil}, pred_intent: {pred_intent}")
        if found_pred_in_avil:
            params["taskgraph"]["intent"] = pred_intent
            for edge in self.graph.out_edges(curr_node, data="intent"):
                if edge[2] == pred_intent:
                    next_node = edge[1]  # found intent under the current node
                    break
            logger.info(f"curr_node: {next_node}")
            node_info, params = self._get_node(next_node, params, intent=pred_intent)
            return True, node_info, params
        return False, {}, params
    
    def handle_unknown_intent(self, curr_node, params: Params) -> Tuple[dict, Params]:
        """
        If unknown intent, call planner
        """
        # if none of the available intents can represent user's utterance, transfer to the planner to let it decide for the next step
        params["taskgraph"]["intent"] = None
        if params["taskgraph"]["nlu_records"]:
            params["taskgraph"]["nlu_records"][-1]["no_intent"] = True  # no intent found
        else:
            params["taskgraph"]["nlu_records"].append({"candidate_intents": [], "pred_intent": "", "no_intent": True, "global_intent": False})
        params["curr_node"] = curr_node
        node_info = NodeInfo(
            resource_id = "planner",
            resource_name = "planner",
            can_skipped=False,
            is_leaf=len(list(self.graph.successors(curr_node))) == 0,
            attributes = {"value": "", "direct": False}
        )
        return node_info, params
        


    def get_node(self, inputs):
        """
        Get the next node
        """
        self.text = inputs["text"]
        self.chat_history_str = inputs["chat_history_str"]
        params: Params = inputs["parameters"]
        # boolean to check if we allow global intent switch or not.
        allow_global_intent_switch = inputs["allow_global_intent_switch"]
        params["taskgraph"]["nlu_records"] = []

        curr_node, params = self.get_current_node(params)
        logger.info(f"Intial curr_node: {curr_node}")

        # For the multi-step nodes, directly stay at that node instead of moving to other nodes
        is_multi_step_node, node_output, params = self.handle_multi_step_node(curr_node, params)
        if is_multi_step_node:
            return node_output, params
        
        if not list(self.graph.successors(curr_node)):  # leaf node
            last_flow_stack_node = self.get_last_flow_stack_node(params)
            if last_flow_stack_node:
                curr_node = last_flow_stack_node
            if self.initial_node:
                curr_node = self.initial_node
        
        # store current node
        params["taskgraph"]["curr_node"] = curr_node
        logger.info(f"curr_node: {curr_node}")

        # available global intents
        available_global_intents = self.get_available_global_intents(params)
        
        # update limit
        params = self.update_node_limit(params)

        # Get local intents of the curr_node
        curr_local_intents = self.get_local_intent(curr_node, params)

        if not curr_local_intents and allow_global_intent_switch:  # no local intent under the current node
            logger.info(f"no local intent under the current node")
            is_global_intent_found, _, node_output, params = \
                self.global_intent_prediction(
                    curr_node,
                    params,
                    available_global_intents,
                    {}
                )
            if is_global_intent_found:
                return node_output, params

        # if current node is incompleted -> return current node
        is_incomplete_node, node_output, params = self.handle_incomplete_node(curr_node, params)
        if is_incomplete_node:
            return node_output, params
        
        # if completed and no local intents -> randomly choose one of the next connected nodes (edges with intent = None)
        if not curr_local_intents:
            logger.info(f"no local or global intent found, move to the next connected node(s)")
            has_random_next_node, node_output, params = self.handle_random_next_node(curr_node, params)
            if has_random_next_node:
                return node_output, params

        logger.info("Finish global condition, start local intent prediction")
        is_local_intent_found, node_output, params = self.local_intent_prediction(curr_node, params, curr_local_intents)
        if is_local_intent_found:
            return node_output, params
        
        pred_intent = None
        if allow_global_intent_switch:
            is_global_intent_found, pred_intent, node_output, params = \
                    self.global_intent_prediction(
                        curr_node,
                        params,
                        available_global_intents,
                        {**curr_local_intents, **{"none": None}}
                    )
            if is_global_intent_found: 
                return node_output, params
        if pred_intent and pred_intent != self.unsure_intent.get("intent"): # if not unsure intent
            # If user didn't indicate all the intent of children nodes under the current node, 
            # then we could randomly choose one of Nones to continue the dialog flow
            has_random_next_node, node_output, params = self.handle_random_next_node(curr_node, params)
            if has_random_next_node:
                return node_output, params
            
        # if none of the available intents can represent user's utterance or it is an unsure intents,
        # transfer to the planner to let it decide for the next step
        node_output, params = self.handle_unknown_intent(curr_node, params)
        return node_output, params
        

    def postprocess_node(self, node) -> Tuple[NodeInfo, Params]:
        node_info: NodeInfo = node[0]
        params: Params = node[1]
        dialog_states = params["taskgraph"].get("dialog_states", {})
        # update the dialog states
        if dialog_states.get(node_info["resource_id"]):
            dialog_states = self.slotfillapi.execute(
                dialog_states.get(node_info["resource_id"]),
                format_chat_history(params["memory"].get("history"))
            )
        params["taskgraph"]["dialog_states"] = dialog_states

        return node_info, params
